package com.sd.business.service.work.impl;

import cn.hutool.core.util.ObjectUtil;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.google.common.collect.HashBasedTable;
import com.ruoyi.common.constant.StatusConstant;
import com.ruoyi.common.core.domain.BaseIdPo;
import com.ruoyi.common.exception.ServiceException;
import com.sd.business.entity.order.po.OrderSku;
import com.sd.business.entity.sku.po.SkuSpec;
import com.sd.business.entity.work.constant.MaterialsConstant;
import com.sd.business.entity.work.dto.WorkOrderSelectDto;
import com.sd.business.entity.work.enums.WorkOrderFixationSpecEnum;
import com.sd.business.entity.work.po.WorkOrder;
import com.sd.business.entity.work.po.WorkOrderDetail;
import com.sd.business.entity.work.vo.WorkOrderVo;
import com.sd.business.mapper.work.WorkOrderMapper;
import com.sd.business.service.order.OrderSkuService;
import com.sd.business.service.sku.SkuSpecService;
import com.sd.business.service.work.WorkOrderDetailService;
import com.sd.business.service.work.WorkOrderService;
import com.sd.business.service.work.impl.fixation.FixationFactory;
import com.sd.business.service.work.impl.fixation.FixationStrategy;
import com.sd.business.service.work.impl.fixation.TypeSettingBo;
import com.sd.business.util.CodeEnum;
import com.sd.business.util.packing.GA;
import com.sd.business.util.packing.entity.PlaceItem;
import com.sd.business.util.packing.entity.Solution;
import com.sd.business.util.packing.model.Item;
import com.sd.framework.util.Assert;
import com.sd.framework.util.TransactionUtil;
import com.sd.framework.util.sql.Sql;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.locks.ReentrantLock;
import java.util.stream.Collectors;


/**
 * <p>
 * 生产工单 服务实现类
 * </p>
 *
 * @author
 * @since 2024-01-02
 */
@Service
public class WorkOrderServiceImpl extends ServiceImpl<WorkOrderMapper, WorkOrder> implements WorkOrderService {

    private static final ReentrantLock schedulingLock = new ReentrantLock();

    @Autowired
    private SkuSpecService skuSpecService;

    @Autowired
    private OrderSkuService orderSkuService;

    @Autowired
    private WorkOrderDetailService workOrderDetailService;

    @Override
    public Page<WorkOrderVo> getPage(WorkOrderSelectDto dto) {

        Page<WorkOrderVo> page = Sql.create(WorkOrderVo.class)
                .selectAll(WorkOrder.class)
                .from(WorkOrder.class)
                .orderByDesc(WorkOrder::getId)
                .page(dto);

        return page;
    }

    @Override
    public WorkOrderVo detail(Long id) {

        WorkOrderVo vo = Sql.create(WorkOrderVo.class)
                .selectAll(WorkOrder.class)
                .from(WorkOrder.class)
                .eq(WorkOrder::getId, id)
                .one();

        Assert.notNull(vo, "未知数据");

        return vo;
    }

    @Override
    public void scheduling() {
        boolean b = schedulingLock.tryLock();

        if (!b) {
            throw new ServiceException("正在排版中，请稍后刷新数据");
        }

        try {

            // 查找未排版完成的订单sku
            List<OrderSku> list = orderSkuService.list(q -> q.eq(OrderSku::getCompleteScheduling, StatusConstant.NO));

            if (list.isEmpty()) {
                return;
            }

            // 根据主材和打印纸区分订单sku
            HashBasedTable<Long, Long, List<OrderSku>> map = HashBasedTable.create();
            for (OrderSku orderSku : list) {
                List<OrderSku> itemList = map.get(orderSku.getBomSpecId(), orderSku.getPrintingPaperBomSpecId());
                if (itemList == null) {
                    itemList = new ArrayList<>();
                    map.put(orderSku.getBomSpecId(), orderSku.getPrintingPaperBomSpecId(), itemList);
                }
                itemList.add(orderSku);
            }

            // 排版
            for (Long bomSpecId : map.rowKeySet()) {
                for (Long printingPaperBomSpecId : map.columnKeySet()) {
                    List<OrderSku> orderSkuList = map.get(bomSpecId, printingPaperBomSpecId);
                    scheduling(bomSpecId, printingPaperBomSpecId, orderSkuList);
                }
            }

        } catch (Exception e) {
            log.error("排版业务发生异常", e);
            throw new ServiceException("排版业务发生异常");
        } finally {
            schedulingLock.unlock();
        }
    }

    /**
     * 排版
     *
     * @param bomSpecId              主材规格id
     * @param printingPaperBomSpecId 打印纸规格id
     * @param list                   排版列表
     */
    private void scheduling(Long bomSpecId, Long printingPaperBomSpecId, List<OrderSku> list) {

        if (ObjectUtil.isNotEmpty(list)) {
            return;
        }

        // 固定排版订单
        List<OrderSku> fixationList = new ArrayList<>();

        // 不固定排版订单
        List<OrderSku> notFixationList = new ArrayList<>();

        // 获取需要排版的sku规格
        List<Long> skuSpecIdList = list.stream().map(OrderSku::getSkuSpecId).collect(Collectors.toList());
        Map<Long, SkuSpec> skuSpecMap = skuSpecService.mapKEntity(BaseIdPo::getId, q -> q.in(BaseIdPo::getId, skuSpecIdList));

        for (OrderSku orderSku : list) {
            Long skuSpecId = orderSku.getSkuSpecId();
            SkuSpec skuSpec = skuSpecMap.get(skuSpecId);
            boolean fixation = WorkOrderFixationSpecEnum.isFixation(skuSpec.getLength(), skuSpec.getWidth());

            if (fixation) {
                fixationList.add(orderSku);
            } else {
                notFixationList.add(orderSku);
            }

        }

        // 固定排版
        fixationScheduling(bomSpecId, printingPaperBomSpecId, skuSpecMap, fixationList);

        // 非固定排版
        notFixationScheduling(bomSpecId, printingPaperBomSpecId, skuSpecMap, notFixationList);

    }

    /**
     * 固定排版
     */
    private void fixationScheduling(Long bomSpecId,
                                    Long printingPaperBomSpecId,
                                    Map<Long, SkuSpec> skuSpecMap,
                                    List<OrderSku> list) {

        if (ObjectUtil.isNotEmpty(list)) {
            return;
        }

        // 根据固定排版规则区分
        Map<WorkOrderFixationSpecEnum, List<OrderSku>> map = list.stream().collect(Collectors.groupingBy(
                item -> {
                    SkuSpec skuSpec = skuSpecMap.get(item.getSkuSpecId());
                    if (skuSpec == null) {
                        throw new ServiceException("未知Sku规格");
                    }
                    if (skuSpec.getLength() == null) {
                        throw new ServiceException("品号为【" + skuSpec.getCode() + "】的sku未定义长度");
                    }
                    if (skuSpec.getWidth() == null) {
                        throw new ServiceException("品号为【" + skuSpec.getCode() + "】的sku未定义宽度");
                    }
                    return WorkOrderFixationSpecEnum.getEnum(skuSpec.getLength(), skuSpec.getWidth());
                }
        ));

        map.forEach(((k, v) -> doFixationScheduling(bomSpecId, printingPaperBomSpecId, k, v)));

    }

    /**
     * 非固定排版
     */
    private void notFixationScheduling(Long bomSpecId,
                                       Long printingPaperBomSpecId,
                                       Map<Long, SkuSpec> skuSpecMap,
                                       List<OrderSku> list) {

        if (ObjectUtil.isNotEmpty(list)) {
            return;
        }

        Item[] items = list.stream().flatMap(item -> {
            Long skuSpecId = item.getSkuSpecId();
            SkuSpec skuSpec = skuSpecMap.get(skuSpecId);
            if (skuSpec == null) {
                throw new ServiceException("未知sku规格，id：" + skuSpecId);
            }

            int schedulingNum = item.getQuantity().subtract(item.getSchedulingNum()).intValue();

            List<Item> itemList = new ArrayList<>();
            for (int i = 0; i < schedulingNum; i++) {
                Long orderSkuId = item.getId();
                double length = skuSpec.getLength().doubleValue() + MaterialsConstant.RESERVE;
                double width = skuSpec.getWidth().doubleValue() + MaterialsConstant.RESERVE;

                itemList.add(new Item(orderSkuId, length, width));
            }

            return itemList.stream();
        }).toArray(Item[]::new);

        Solution solve = GA.solve(MaterialsConstant.MASTER_LENGTH, MaterialsConstant.MASTER_WIDTH, items);

        List<PlaceItem> placeItemList = solve.getPlaceItemList();
        if (placeItemList.size() == list.size()) {
            return;
        }

        Map<Long, OrderSku> map = list.stream().collect(Collectors.toMap(BaseIdPo::getId, item -> item));

        // 生产工单
        WorkOrder workOrder = new WorkOrder();
        workOrder.setId(IdWorker.getId());
        workOrder.setCode(CodeEnum.WORK_ORDER_CODE.getCode());
        workOrder.setType(2);
        workOrder.setMasterBomSpecId(bomSpecId);
        workOrder.setPrintingPaperBomSpecId(printingPaperBomSpecId);
        workOrder.setSchedulingNum(new BigDecimal(placeItemList.size()));
        workOrder.setUseRatio(BigDecimal.valueOf(solve.getUseRatio() * 100));
        workOrder.setMaxLength(BigDecimal.valueOf(solve.getMaxLength()));
        workOrder.setSyncProduction(StatusConstant.NO);
        workOrder.setIntegrity(StatusConstant.YES);

        // 生产工单明细
        List<WorkOrderDetail> workOrderDetailList = placeItemList.stream().map(item -> {
            OrderSku orderSku = map.get(item.getId());

            BigDecimal schedulingNum = orderSku.getSchedulingNum().add(BigDecimal.ONE);
            orderSku.setSchedulingNum(schedulingNum);
            if (schedulingNum.compareTo(BigDecimal.ZERO) == 0) {
                orderSku.setCompleteScheduling(StatusConstant.YES);
            }

            WorkOrderDetail workOrderDetail = new WorkOrderDetail();
            workOrderDetail.setWorkOrderId(workOrder.getId());
            workOrderDetail.setOrderInfoId(orderSku.getOrderId());
            workOrderDetail.setOrderSkuId(item.getId());
            workOrderDetail.setX(BigDecimal.valueOf(item.getX()));
            workOrderDetail.setY(BigDecimal.valueOf(item.getY()));
            workOrderDetail.setBleedingLength(BigDecimal.valueOf(item.getLength()));
            workOrderDetail.setBleedingWidth(BigDecimal.valueOf(item.getLength()));
            workOrderDetail.setLength(BigDecimal.valueOf(item.getLength() - MaterialsConstant.RESERVE));
            workOrderDetail.setWidth(BigDecimal.valueOf(item.getWidth() - MaterialsConstant.RESERVE));
            workOrderDetail.setRotate(item.isRotate() ? StatusConstant.YES : StatusConstant.NO);
            return workOrderDetail;
        }).collect(Collectors.toList());

        // 保存数据
        TransactionUtil.execute(() -> {
            orderSkuService.updateBatchById(list);
            save(workOrder);
            workOrderDetailService.saveBatch(workOrderDetailList);
        });

    }

    /**
     * 固定排版
     */
    private void doFixationScheduling(Long bomSpecId,
                                      Long printingPaperBomSpecId,
                                      WorkOrderFixationSpecEnum workOrderFixationSpecEnum,
                                      List<OrderSku> list) {

        // 本次所需排版数量
        int schedulingNum = workOrderFixationSpecEnum.getMaxNum();

        // 未排单数量
        int notSchedulingNum = list.stream().map(item -> item.getQuantity().subtract(item.getSchedulingNum()))
                .reduce(BigDecimal.ZERO, BigDecimal::add).intValue();

        // 如果未排版数量小于本次所需排版数量，则不排单
        if (notSchedulingNum < schedulingNum) {
            return;
        }

        FixationStrategy fixationStrategy = FixationFactory.get(workOrderFixationSpecEnum);

        // 生产工单
        WorkOrder workOrder = new WorkOrder();
        workOrder.setId(IdWorker.getId());
        workOrder.setCode(CodeEnum.WORK_ORDER_CODE.getCode());
        workOrder.setType(1);
        workOrder.setFixationSpec(workOrderFixationSpecEnum.getLength() + "*" + workOrderFixationSpecEnum.getWidth());
        workOrder.setMasterBomSpecId(bomSpecId);
        workOrder.setPrintingPaperBomSpecId(printingPaperBomSpecId);
        workOrder.setSyncProduction(StatusConstant.NO);
        workOrder.setIntegrity(StatusConstant.YES);

        // 生产工单明细
        List<WorkOrderDetail> workOrderDetailList = new ArrayList<>();

        for (OrderSku orderSku : list) {

            // 此单未排版数量
            int itemSchedulingNum = orderSku.getQuantity().subtract(orderSku.getSchedulingNum()).intValue();

            // 未排排数量小于本次所需排版数量
            if (schedulingNum >= itemSchedulingNum) {

                // 赋值排版数量等于sku数量
                orderSku.setSchedulingNum(orderSku.getQuantity());
                orderSku.setCompleteScheduling(StatusConstant.YES);

                for (int i = 0; i < itemSchedulingNum; i++) {
                    addWorkOrderDetail(fixationStrategy, workOrder, workOrderDetailList, orderSku);
                }

                // 重新计算需要排版数量
                schedulingNum -= itemSchedulingNum;
            } else {
                orderSku.setSchedulingNum(new BigDecimal(itemSchedulingNum - schedulingNum));
                orderSku.setCompleteScheduling(StatusConstant.NO);

                for (int i = 0; i < schedulingNum; i++) {
                    addWorkOrderDetail(fixationStrategy, workOrder, workOrderDetailList, orderSku);
                }

                break;
            }

        }

        //workOrder.setSchedulingNum();

        // 保存数据
        TransactionUtil.execute(() -> {
            orderSkuService.updateBatchById(list);
            save(workOrder);
            workOrderDetailService.saveBatch(workOrderDetailList);
        });

        // 再次执行判断是否还可以排版
        doFixationScheduling(bomSpecId, printingPaperBomSpecId, workOrderFixationSpecEnum, list);

    }


    private void addWorkOrderDetail(FixationStrategy fixationStrategy, WorkOrder workOrder,
                                    List<WorkOrderDetail> workOrderDetailList, OrderSku orderSku) {

        TypeSettingBo typeSettingBo = fixationStrategy.next();

        WorkOrderDetail workOrderDetail = new WorkOrderDetail();
        workOrderDetail.setWorkOrderId(workOrder.getId());
        workOrderDetail.setOrderInfoId(orderSku.getOrderId());
        workOrderDetail.setOrderSkuId(orderSku.getId());

        workOrderDetail.setLength(typeSettingBo.getLength());
        workOrderDetail.setWidth(typeSettingBo.getWidth());
        workOrderDetail.setBleedingLength(typeSettingBo.getBleedingLength());
        workOrderDetail.setBleedingWidth(typeSettingBo.getBleedingWidth());
        workOrderDetail.setRotate(typeSettingBo.getHasRotate() ? StatusConstant.YES : StatusConstant.NO);
        workOrderDetail.setX(typeSettingBo.getX());
        workOrderDetail.setY(typeSettingBo.getY());

        workOrderDetailList.add(workOrderDetail);
    }


}
